
/******************************************************************************
* MODULE     : ext_generic.cc
* DESCRIPTION: Produce the generic extension code
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
namespace mmx {

/******************************************************************************
* Specializing list of imports to a given type
******************************************************************************/

generic
extender::specialize (generic g, generic to, bool virtual_flag) {
  // cout << "Specialize generic method " << g << " to " << to << lf;
  if (!(is_func (g, MMX_FORALL, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	(g[1][2] == MMX_CLASS) &&
	is_func (g[2], MMX_DEFINE, 2) &&
	is_func (g[2][1], MMX_TYPE, 2)))
    error (g, "Assertion failed in specialize");
  else {
    generic qvar= g[1][1];
    generic body= g[2];
    generic ret = substitute (body, qvar, to);
    if (virtual_flag)
      ret= list (ret[0],
		 list (ret[1][0], ret[1][1], Virtual (ret[1][2])),
		 ret[2]);
    return ret;
  }
  return g;
}

list
extender::specialize (list l, generic to, bool virtual_flag) {
  if (nil (l)) return l;
  else return cons (specialize (car (l), to, virtual_flag),
		    specialize (cdr (l), to, virtual_flag));
}

/******************************************************************************
* Produce the gentype type. This type serves as a base type for
* all dynamic types which have been exported by the user.
******************************************************************************/

static list
arg_list (int start, int end) {
  if (start >= end) return list ();
  else return cons (parameter (start), arg_list (start+1, end));
}

list
extender::gentype_args_as_mmx (list l) {
  if (nil (l)) return l;
  generic var = car (l) [1];
  generic type= car (l) [2];
  generic rval= view_as_mmx (var, type, table ());
  generic rtype= build_mmx_type (type, list ());
  return cons (list (mmx_symbol ("value"), rval, rtype),
	       gentype_args_as_mmx (cdr (l)));
}

void
extender::gentype_methods (generic h) {
  generic g= decl_as_ccpp (h);
  // cout << "Generate method " << g << " for " << gentype << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_symbol (g[1]) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in gentype_methods");
  else {
    g= name_parameters (g);
    generic name= genmethod (g[1]);
    generic scoped= Scope (gentype_rep, name);
    generic fun= list (g[0], scoped, g[2]);
    if (glue_flag) {
      h= name_parameters (h[1]);
      generic back= build_mmx_type (h[2][2], list ());
      back= substitute (back, geninst (), symbol ("mmxt"));
      back= list (mmx_symbol ("value"), back, mmx_symbol ("ID_CLASS"));
      string sname= as_string (S (as_symbol (h[1])));
      generic symb= list (mmx_symbol ("symbol"), sname);
      if (sname == "new") symb= list (mmx_symbol ("MMX_NEW"), back[1]);
      list args= gentype_args_as_mmx (cdr (as_list (h[2][1])));
      generic instr= mmx_make_list (cons (symb, args));
      instr= list (mmx_symbol ("list"), mmx_symbol ("MMX_TRANSTYPE"),
		   instr, back);
      generic eval= list (mmx_symbol ("extern_eval"), instr);
      generic ret= view_as_ccpp (eval, h[2][2], table ());
      generate (Define (fun, Begin (Return (ret))));
    }
    else {
      string err ("\"Fatal error: Generic method " *
		  as_string (S (as_symbol (g[1]))) *
		  " is not supported\\n\"");
      generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
      generic _exit = list (symbol ("exit"), symbol ("1"));
      generate (Define (fun, Begin (_error, _exit)));
    }
  }
}

void
extender::gentype_methods (list l) {
  if (!nil (l)) {
    gentype_methods (car (l));
    gentype_methods (cdr (l));
  }
}

void
extender::produce_gentype () {
  // The gentype is a pointer to gentype_rep
  generate (Type (gentype_rep, MMX_CLASS));
  generate (Typedef (gentype, Pointer (gentype_rep)));

  // Generate the gentype_rep structure
  list l= specialize (gen_imports, geninst (), true);
  l= decl_as_ccpp (l, true);
  generic gt= Type (symbol ("mmxt"), mmx_generic);
  generic gc= Define (Type (gentype_rep, Inline (Constructor (MMX_VOID))),
		      Begin ());
  generic gd= Define (Type (gentype_rep, Inline (Virtual (Destructor ()))),
		      Begin ());
  generic in= Define (Type (symbol ("internal"),
			    Virtual (Function (MMX_VOID, MMX_BOOL))),
		      Begin (Return (symbol ("false"))));
  if (glue_flag) l= cons (gt, gc, gd, in, l);
  else l= cons (gc, gd, in, l);
  generate (make_struct (gentype_rep, l));
  if (glue_flag)
    generate (list (symbol ("GENTYPE_CONVERTERS"), gentype, gentype_rep));

  // Printing geninst instances should never happen,
  // but the routine should be included to make the compiler happy
  if (glue_flag) {
    generic x  = symbol ("g");
    generic out= symbol ("out");
    generic ost= Alias (MMX_OSTREAM);
    generic dom= Cross (Type (out, ost), Type (x, geninst ()));
    generic fun= Type (MMX_LESSLESS, Function (dom, ost));
    string err ("\"Fatal error: << should not be called for generics\"");
    generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
    generic _exit = list (symbol ("exit"), symbol ("1"));
    generate (Define (fun, Begin (_error, _exit)));
  }

  // Generate the methods
  l= specialize (gen_imports, geninst (), false);
  gentype_methods (l);
}

/******************************************************************************
* Produce the unnumbered generic instance types
******************************************************************************/

static generic
converter_class (
  generic tmpl_var, generic cvr, generic F, generic T, generic body)
{
  generic x  = symbol ("x");
  generic h  = symbol ("helper");

  // the contents of the converter structure
  generic tx = Type (x, F);
  generic f  = Static (Inline (Function (tx, T)));
  generic b  = Define (Type (h, f), body);
  generic val= Struct (b);

  // the converter class with helper routine
  generic var= Specialize (cvr, T, F);
  generic tp = Forall (Type (tmpl_var, MMX_CLASS), MMX_CLASS);
  if (N(tmpl_var) == 2) {
    if ((F == symbol ("F")) && (T == symbol ("T"))) var= cvr;
    tp = Forall (Type (tmpl_var[0], MMX_CLASS),
		 Forall (Type (tmpl_var[1], MMX_CLASS),
			 MMX_CLASS));
  }
  return Define (Type (var, tp), val);
}

void
extender::produce_geninst () {
  generate (make_blackbox_struct (geninst ()));
  if (glue_flag)
    generate (list (symbol ("GENINST_CONVERTERS"), geninst ()));

  generic x  = symbol ("x");
  generic h  = symbol ("helper");
  generic cv = symbol ("convert");
  generic cvr= symbol ("converter");
  generic Cv = symbol ("cast");
  generic Cvr= symbol ("caster");
  generic F  = symbol ("F");
  generic T  = symbol ("T");

  // The converter meta template
  generic b0= Begin (Return (Transtype (x, T)));
  generate (converter_class (list (T, F), cvr, F, T, b0));

  // From abstract to concrete
  generic a1= Transtype (Dot (x, symbol ("rep")), Pointer (T));
  generic b1= Begin (Return (Unpointer (Brackets (a1))));
  generate (converter_class (T, cvr, geninst (), T, b1));

  // From concrete to abstract
  generic b2= Begin (Return (list (geninst (), New (F, x))));
  generate (converter_class (F, cvr, F, geninst (), b2));

  // From abstract aliases to concrete aliases
  generate (converter_class (T, cvr, Alias (geninst ()), Alias (T), b1));

  // From concrete aliases to abstract aliases
  generic a4= New (geninst (), Unalias (x));
  generic b4= Begin (Return (Unpointer (a4)));
  generate (converter_class (F, cvr, Alias (F), Alias (geninst ()), b4));

  // The main conversion routine
  generic cc= Scope (list (cvr, T, F), h);
  generic ff= Inline (Function (Type (x, F), T));
  generic tt= Forall (Type (T, MMX_CLASS), Forall (Type (F, MMX_CLASS), ff));
  generic bb= Return (list (cc, x));
  generate (Define (Type (cv, tt), Begin (bb)));

  // Crude casting meta template
  generic B0= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, F, T, B0));

  // Crude casting for aliases meta template
  generic B1= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, Alias (F), Alias (T), B1));

  // The main conversion routine for crude casting
  generic Cc= Scope (list (Cvr, T, F), h);
  generic Bb= Return (list (Cc, x));
  generate (Define (Type (Cv, tt), Begin (Bb)));
}

/******************************************************************************
* Produce the generic instance types, indexed by numbers 1, 2, ....
* These types are instances of dynamic types. curinst(i) holds
* the current type of geninst(i).
******************************************************************************/

list
extender::abstract (int nr, list args, list types) {
  // Convert to abstract type when necessary
  if (nil (args)) return args;
  generic x= car (args);
  generic stype= car (types) [2];
  generic dtype= substitute (stype, geninst (nr), geninst ());
  if (dtype != stype)
    x= list (Specialize (symbol ("cast"), dtype, stype), x);
  return cons (x, abstract (nr, cdr (args), cdr (types)));
}

void
extender::geninst_constructors (int nr, generic g, list& r) {
  // cout << "Generate constructor " << g << " for " << geninst (nr) << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in geninst_constructors");
  else if (g[1] != symbol ("new"));
  else {
    bool copy_flag= (g[2][1] == Cross (g[2][2]));
    g= name_parameters (g);
    generic rep = symbol ("rep");
    generic name= genmethod (g[1]);
    generic fun = Arrow (curtype (1), name);
    list    args= arg_list (1, N(g[2][1]));
    generic appl= cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    if (copy_flag)
      appl= list (fun, list (geninst(), Dot (parameter (1), rep)));
    generic inst= Dot (appl, rep);
    generic body= Initialize (generic (list (rep, inst)), Begin ());
    list    tps = cdr (as_list (g[2][1]));
    if (copy_flag)
      tps= list (Type (tps[0][1], Const (Alias (tps[0][2]))));
    generic type= Inline (Constructor (tps));
    r= cons (Define (list (g[0], geninst (nr), type), body), r);

    if (copy_flag) {
      list    tps = cdr (as_list (g[2][1]));
      generic cmd1= Assign (rep, inst);
      generic cmd2= Return (Unpointer (symbol ("this")));
      generic body= Begin (cmd1, cmd2);
      generic type= Inline (Function (Cross (tps), Alias (g[2][2])));
      r= cons (Define (Type (symbol ("operator ="), type), body), r);
    }
  }
}

list
extender::geninst_constructors (int nr, list l) {
  list r;
  while (!nil (l)) {
    geninst_constructors (nr, car (l), r);
    l= cdr (l);
  }
  return reverse (r);
}

void
extender::geninst_methods (int nr, generic g) {
  // cout << "Generate method " << g << " for " << geninst (nr) << lf;
  if (!(is_func (g, MMX_TYPE, 2) &&
	is_symbol (g[1]) &&
	is_func (g[2], MMX_FUNCTION, 2) &&
	is_func (g[2][1], MMX_CROSS)))
    error (g, "Assertion failed in geninst_methods");
  else if (g[1] == symbol ("new"));
  else {
    g= name_parameters (g);
    generic name = genmethod (g[1]);
    generic fun  = Arrow (curtype (1), name);
    list    args = arg_list (1, N(g[2][1]));
    generic appl = cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    generic dtype= g[2][2];
    generic stype= substitute (dtype, geninst (nr), geninst ());
    if (dtype != stype)
      appl= list (Specialize (symbol ("cast"), dtype, stype), appl);
    if (dtype == MMX_VOID) generate (Define (g, Begin (appl)));
    else generate (Define (g, Begin (Return (appl))));
  }
}

void
extender::geninst_methods (int nr, list l) {
  if (!nil (l)) {
    geninst_methods (nr, car (l));
    geninst_methods (nr, cdr (l));
  }
}

void
extender::produce_geninst (int nr) {
  generate (Type (curtype (nr), Static (gentype)));
  list l = specialize (gen_imports, geninst (nr), false);
  l      = decl_as_ccpp (l);
  list cl= geninst_constructors (nr, l);
  generate (make_blackbox_struct (geninst (nr), cl));
  if (glue_flag)
    generate (list (symbol ("GENINSTX_CONVERTERS"),
		    geninst (nr), curtype (nr)));
  geninst_methods (nr, l);
}

} // namespace mmx
