
/******************************************************************************
* MODULE     : quantor.cc
* DESCRIPTION: quantified variables, identified by unique numbers
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <quantor.hh>
namespace mmx {

long long int qt_serial;

generic quantor_normal_form (generic g, int& i);

/******************************************************************************
* Required quantors for generic structures
******************************************************************************/

int
quantor_rep::Type () {
  return QUANTOR_TYPE;
}

int
quantor_rep::Hash () {
  return (int) nr;
}

bool
quantor_rep::Equal (generic g) {
  if (gentype (g) != QUANTOR_TYPE) return false;
  return quantor (this) == as_quantor (g);
}

generic
quantor_rep::Copy () {
  return copy (quantor (this));
}

int
quantor_rep::Arity () {
  return 0;
}

void
quantor_rep::Print (ostream& out) {
  out << quantor (this);
}

/******************************************************************************
* Routines for quantors
******************************************************************************/

ostream&
operator << (ostream& out, quantor v) {
  if (v->type == qt_free) return out << "?[" << v->nr << "]";
  if (v->type == qt_witness) return out << "![" << v->nr << "]";
  return out << v->name;
}

quantor
copy (quantor v) {
  return quantor (v->type, v->nr, v->name);
}

generic
highest_in (generic f) {
  if (is_quantor (f) && (!is_bound_var (f))) return f;
  if (!is_list (f)) return generic ();
  list l= as_list (f);
  generic r;
  while (!nil (l)) {
    generic s= highest_in (car (l));
    if (!nil (s))
      if (nil (r) || (as_quantor(s)->nr > as_quantor(r)->nr))
	r= s;
    l= cdr (l);
  }
  return r;
}

} // namespace mmx
