
/******************************************************************************
* MODULE     : list.cc
* DESCRIPTION: lists
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <list.hh>
namespace mmx {

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

int
list_rep::Type () {
  return LIST_TYPE;
}

int
list_rep::Hash () {
  register int h= hash (item);
  return (h<<7) + (h>>25) + hash (next);
}

bool
list_rep::Equal (generic g) {
  if (gentype (g) != LIST_TYPE) return false;
  return list (this) == as_list (g);
}

generic
list_rep::Copy () {
  return copy (list (this));
}

int
list_rep::Arity () {
  return N (list (this));
}

generic&
list_rep::Access (int i) {
  return list (this) [i];
}

generic
list_rep::Range (int begin, int end) {
  return generic (list (this) (begin, end));
}

generic&
list_rep::Car () {
  return item;
}

generic
list_rep::Cdr () {
  return generic (next);
}

void
list_rep::Print (ostream& out) {
  out << list (this);
}

/******************************************************************************
* Standard routines for lists
******************************************************************************/

int
N (list l) {
  int n=0;
  while (!nil (l)) {
    n++;
    l= cdr (l);
  }
  return n;
}

bool
operator == (list l1, list l2) {
  while (!nil (l1)) {
    if (nil (l2)) return false;
    if (car (l1) != car (l2)) return false;
    l1= cdr (l1);
    l2= cdr (l2);
  }
  return nil (l2);
}

bool
operator != (list l1, list l2) {
  while (!nil (l1)) {
    if (nil (l2)) return true;
    if (car (l1) != car (l2)) return true;
    l1= cdr (l1);
    l2= cdr (l2);
  }
  return !nil (l2);
}

list
copy (list l) {
  if (nil (l)) return l;
  return cons (car (l), copy (cdr (l)));
}

generic&
list::operator [] (int i) {
  if (i==0) return rep->item;
  else return list (rep->next) [i-1];
}

list
list::operator () (int start, int end) {
  if (start == 0) {
    if (end == 0) return list ();
    else return cons (rep->item, list (rep->next) (0, end-1));
  }
  else return list (rep->next) (start-1, end-1);
}

ostream&
operator << (ostream& out, list l) {
  print_entries (out, l, true);
  return out;
}

/******************************************************************************
* Other routines for lists
******************************************************************************/

list
reverse (list l) {
  list r;
  while (!nil(l)) {
    r= cons (car(l), r);
    l= cdr(l);
  }
  return r;
}

list
operator * (list l1, list l2) {
  if (nil (l1)) return l2;
  list r (car (l1));
  list lc= r;
  while (!nil (cdr (l1))) {
    l1= cdr (l1);
    cdr (lc)= list (car (l1));
    lc= cdr (lc);
  }
  cdr (lc)= l2;
  return r;
}

list
apply_all (generic g, list l) {
  if (nil (l)) return l;
  else return cons (list (g, car (l)), apply_all (g, cdr (l)));
}

bool
is_list (generic g, generic g1) {
  return (!nil (g)) && (g->Type() == LIST_TYPE) && (car (g) == g1);
}

bool
is_list (generic g, generic g1, generic g2) {
  return is_list (g, g1) && is_list (cdr (g), g2);
}

} // namespace mmx
